# Size of puzzle. Demo is using 4x4 tiles.
define block_shift_game_rows = 4
define block_shift_game_columns = 4

# How many moves should be allowed before the game is forfeit?
define block_shift_game_max_moves = 300

# How throuhly should the game shuffle the puzzle?
define block_shift_game_initial_shuffle = 2000

# Where to look for the tiles. First {} get's replaced by a letter indicating the row. Second is the column#.
define block_shift_game_pattern = "blocks/block-{}{}.png"

# If you'd like a little space between your tiles, play with this parameter.
# Note: If you cut up an image to be used as blocks you might either want to set the spacing to 0 or shave a few pixels off the borders of the pieces.
define block_shift_game_spacing = 2

# Style properties for the block_shift_game_screen's UI elements.
# See https://www.renpy.org/dev-doc/html/screens.html#style-prefix
style block_shift_game_vbox:
    xalign 0.5
    yalign 0.0

style block_shift_game_grid:
    spacing block_shift_game_spacing

style block_shift_game_button:
    xalign 0.5

style block_shift_game_button_text:
    size 40
    bold True


init python:
    class Block (object):

        def __init__ (self, image=None, home_x=None, home_y=None):
            self.image = image
            self.home_x = home_x
            self.home_y = home_y

        def __str__ (self):
            return self.image if self else "--"

        def __repr__ (self):
            return "{}({},{})".format (self.image, self.home_x, self.home_y) if self else "--"

        def __nonzero__ (self):
            return self.image is not None and self.home_x is not None and self.home_y is not None

    spacer = Block ()

    class Board (object):

        def __init__ (self, image_pattern, rows, columns):
            self.content = [[None]*(2+columns)]
            self.empty_x = columns
            self.empty_y = rows
            self.rows = rows
            self.columns = columns
            for y in range (1, 1+rows):
                row = [None]
                for x in range (1, 1+columns):
                    if y == self.empty_y and x == self.empty_x:
                        row.append (spacer)
                    else:
                        row.append (Block (image_pattern.format (chr(64+y), x), x, y))
                row.append (None)
                self.content.append (row)
            self.content.append ([None]*(2+columns))

        def __str__ (self):
            ret = []
            for row in self.content:
                ret.append (row.__str__ ())
            return "||".join (ret)

        def image_for (self, x, y):
            return self.content [y+1][x+1].image if self.content [y+1][x+1] else None

        def classify_at (self, x, y):
            x += 1
            y += 1
            if not self.content [y][x]:
                return None
            elif x == self.empty_x-1 and y == self.empty_y  :
                return "left"
            elif x == self.empty_x+1 and y == self.empty_y  :
                return "right"
            elif x == self.empty_x   and y == self.empty_y-1:
                return "up"
            elif x == self.empty_x   and y == self.empty_y+1:
                return "down"
            else:
                return "image"

        def is_solved (self):
            for y in range (1, 1+self.rows):
                for x in range (1, 1+self.columns):
                    if self.content [y][x]:
                        if self.content [y][x].home_x != x or self.content [y][x].home_y != y:
                            return False
            return True

        def show_spacer (self):
            print self.content [self.empty_y][self.empty_x], self.empty_y, self.empty_x

        def swap (self, x1, y1, x2, y2):
            if isinstance (self.content [y1][x1], Block) and isinstance (self.content [y2][x2], Block):
                tmp = self.content [y1][x1]
                self.content [y1][x1] = self.content [y2][x2]
                self.content [y2][x2] = tmp
                return True
            return False

        def move_spacer_left (self):
            if self.swap (self.empty_x, self.empty_y, self.empty_x-1, self.empty_y  ):
                self.empty_x -= 1
            # self.show_spacer ()

        def move_spacer_right (self):
            if self.swap (self.empty_x, self.empty_y, self.empty_x+1, self.empty_y  ):
                self.empty_x += 1
            # self.show_spacer ()

        def move_spacer_up (self):
            if self.swap (self.empty_x, self.empty_y, self.empty_x  , self.empty_y-1):
                self.empty_y -= 1
            # self.show_spacer ()

        def move_spacer_down (self):
            if self.swap (self.empty_x, self.empty_y, self.empty_x  , self.empty_y+1):
                self.empty_y += 1
            # self.show_spacer ()

        def shuffle (self, count=1):
            for i in range (count):
                what = renpy.random.randint (1, 4)
                if what == 1:
                    self.move_spacer_left ()
                elif what == 2:
                    self.move_spacer_right ()
                elif what == 3:
                    self.move_spacer_up ()
                else:
                    self.move_spacer_down ()

screen block_shift_game_screen:
    style_prefix "block_shift_game"
    vbox:
        grid block_shift_game_rows block_shift_game_columns:
            for y in range (block_shift_game_rows):
                for x in range (block_shift_game_columns):
                    $ block_shift_game_type = block_shift_game_board.classify_at (x, y)
                    if block_shift_game_type == "image":
                        add block_shift_game_board.image_for (x, y)
                    elif block_shift_game_type is None:
                        null
                    else:
                        imagebutton idle block_shift_game_board.image_for (x, y) action Return (block_shift_game_type)

        textbutton "Give Up!" action Return ("quit")

label block_shift_game:
    $ block_shift_game_board = Board (image_pattern=block_shift_game_pattern, rows=block_shift_game_rows, columns=block_shift_game_columns)
    $ block_shift_game_board.shuffle (block_shift_game_initial_shuffle)
    $ moves = 0

    scene black

    while not block_shift_game_board.is_solved () and not moves > block_shift_game_max_moves:
        call screen block_shift_game_screen
        $ rv = _return
        if rv == "left":
            $ block_shift_game_board.move_spacer_left ()
            $ moves += 1
        if rv == "right":
            $ block_shift_game_board.move_spacer_right ()
            $ moves += 1
        if rv == "up":
            $ block_shift_game_board.move_spacer_up ()
            $ moves += 1
        if rv == "down":
            $ block_shift_game_board.move_spacer_down ()
            $ moves += 1
        if rv == "quit":
            return 0

    return moves if moves <= block_shift_game_max_moves else 0
